@@ -7,7 +7,10 @@
 /* Sharable code and data for wrapper around JSObjects. */
 
 #include "xpcprivate.h"
+#include "js/CallAndConstruct.h"  // JS_CallFunctionValue
+#include "js/Object.h"            // JS::GetClass
 #include "js/Printf.h"
+#include "js/PropertyAndElement.h"  // JS_Enumerate, JS_GetProperty, JS_GetPropertyById, JS_HasProperty, JS_HasPropertyById, JS_SetProperty, JS_SetPropertyById
 #include "nsArrayEnumerator.h"
 #include "nsINamed.h"
 #include "nsIScriptError.h"
@@ -16,12 +19,11 @@
 #include "nsJSUtils.h"
 #include "nsPrintfCString.h"
 #include "mozilla/Attributes.h"
+#include "mozilla/dom/AutoEntryScript.h"
 #include "mozilla/dom/BindingUtils.h"
 #include "mozilla/dom/DOMException.h"
 #include "mozilla/dom/DOMExceptionBinding.h"
-#include "mozilla/dom/DOMPrefs.h"
 #include "mozilla/dom/MozQueryInterface.h"
-#include "mozilla/jsipc/CrossProcessObjectWrappers.h"
 
 #include "jsapi.h"
 #include "jsfriendapi.h"
@@ -98,11 +100,7 @@ class MOZ_STACK_CLASS AutoSavePendingResult {
 // static
 const nsXPTInterfaceInfo* nsXPCWrappedJS::GetInterfaceInfo(REFNSIID aIID) {
   const nsXPTInterfaceInfo* info = nsXPTInterfaceInfo::ByIID(aIID);
-  if (!info) {
-    return nullptr;
-  }
-
-  if (info->IsBuiltinClass() || !nsXPConnect::IsISupportsDescendant(info)) {
+  if (!info || info->IsBuiltinClass()) {
     return nullptr;
   }
 
@@ -150,14 +148,6 @@ JSObject* nsXPCWrappedJS::CallQueryInterfaceOnJSObject(JSContext* cx,
     return nullptr;
   }
 
-  // Ensure that we are asking for a non-builtinclass interface
-  if (!aIID.Equals(NS_GET_IID(nsISupports))) {
-    const nsXPTInterfaceInfo* info = nsXPTInterfaceInfo::ByIID(aIID);
-    if (!info || info->IsBuiltinClass()) {
-      return nullptr;
-    }
-  }
-
   dom::MozQueryInterface* mozQI = nullptr;
   if (NS_SUCCEEDED(UNWRAP_OBJECT(MozQueryInterface, &fun, mozQI))) {
     if (mozQI->QueriesTo(aIID)) {
@@ -220,7 +210,7 @@ namespace {
 class WrappedJSNamed final : public nsINamed {
   nsCString mName;
 
-  ~WrappedJSNamed() {}
+  ~WrappedJSNamed() = default;
 
  public:
   NS_DECL_ISUPPORTS
@@ -315,12 +305,27 @@ nsresult nsXPCWrappedJS::DelegatedQueryInterface(REFNSIID aIID,
     return NS_OK;
   }
 
-  // We can't have a cached wrapper.
-  if (aIID.Equals(NS_GET_IID(nsWrapperCache))) {
-    *aInstancePtr = nullptr;
-    return NS_NOINTERFACE;
+  // Ensure that we are asking for a non-builtinclass interface, and avoid even
+  // setting up our AutoEntryScript if we are.  Don't bother doing that check
+  // if our IID is nsISupports: we know that's not builtinclass, and we QI to
+  // it a _lot_.
+  if (!aIID.Equals(NS_GET_IID(nsISupports))) {
+    const nsXPTInterfaceInfo* info = nsXPTInterfaceInfo::ByIID(aIID);
+    if (!info || info->IsBuiltinClass()) {
+      MOZ_ASSERT(!aIID.Equals(NS_GET_IID(nsISupportsWeakReference)),
+                 "Later code for nsISupportsWeakReference is being skipped");
+      MOZ_ASSERT(!aIID.Equals(NS_GET_IID(nsISimpleEnumerator)),
+                 "Later code for nsISimpleEnumerator is being skipped");
+      MOZ_ASSERT(!aIID.Equals(NS_GET_IID(nsINamed)),
+                 "Later code for nsINamed is being skipped");
+      *aInstancePtr = nullptr;
+      return NS_NOINTERFACE;
+    }
   }
 
+  MOZ_ASSERT(!aIID.Equals(NS_GET_IID(nsWrapperCache)),
+             "Where did we get non-builtinclass interface info for this??");
+
   // QI on an XPCWrappedJS can run script, so we need an AutoEntryScript.
   // This is inherently Gecko-specific.
   // We check both nativeGlobal and nativeGlobal->GetGlobalJSObject() even
@@ -330,6 +335,9 @@ nsresult nsXPCWrappedJS::DelegatedQueryInterface(REFNSIID aIID,
   nsIGlobalObject* nativeGlobal = NativeGlobal(js::UncheckedUnwrap(obj));
   NS_ENSURE_TRUE(nativeGlobal, NS_ERROR_FAILURE);
   NS_ENSURE_TRUE(nativeGlobal->HasJSGlobal(), NS_ERROR_FAILURE);
+
+  AutoAllowLegacyScriptExecution exemption;
+
   AutoEntryScript aes(nativeGlobal, "XPCWrappedJS QueryInterface",
                       /* aIsMainThread = */ true);
   XPCCallContext ccx(aes.cx());
@@ -556,6 +564,11 @@ void nsXPCWrappedJS::CleanupOutparams(const nsXPTMethodInfo* info,
 
     MOZ_ASSERT(param.IsIndirect(), "Outparams are always indirect");
 
+    // Don't try to clear optional out params that are not set.
+    if (param.IsOptional() && !nativeParams[i].val.p) {
+      continue;
+    }
+
     // Call 'CleanupValue' on parameters which we know to be initialized:
     //  1. Complex parameters (initialized by caller)
     //  2. 'inout' parameters (initialized by caller)
@@ -600,7 +613,7 @@ nsresult nsXPCWrappedJS::CheckForException(XPCCallContext& ccx,
   RootedValue js_exception(cx);
   bool is_js_exception = JS_GetPendingException(cx, &js_exception);
 
-  /* JS might throw an expection whether the reporter was called or not */
+  /* JS might throw an exception whether the reporter was called or not */
   if (is_js_exception) {
     if (!xpc_exception) {
       XPCConvert::JSValToXPCException(cx, &js_exception, anInterfaceName,
@@ -657,7 +670,7 @@ nsresult nsXPCWrappedJS::CheckForException(XPCCallContext& ccx,
     }
 
     if (reportable) {
-      if (DOMPrefs::DumpEnabled()) {
+      if (nsJSUtils::DumpEnabled()) {
         static const char line[] =
             "************************************************************\n";
         static const char preamble[] =
@@ -707,7 +720,7 @@ nsresult nsXPCWrappedJS::CheckForException(XPCCallContext& ccx,
             }
 
             nsresult rv = scriptError->InitWithWindowID(
-                NS_ConvertUTF8toUTF16(newMessage), sourceName, EmptyString(),
+                NS_ConvertUTF8toUTF16(newMessage), sourceName, u""_ns,
                 lineNumber, 0, 0, "XPConnect JavaScript",
                 nsJSUtils::GetCurrentlyRunningCodeInnerWindowID(cx));
             if (NS_FAILED(rv)) {
@@ -751,6 +764,14 @@ nsXPCWrappedJS::CallMethod(uint16_t methodIndex, const nsXPTMethodInfo* info,
     return NS_ERROR_UNEXPECTED;
   }
 
+  // We need to reject an attempt to call a non-reflectable method before
+  // we do anything like AutoEntryScript which might allocate in the JS engine,
+  // because the method isn't marked with JS_HAZ_CAN_RUN_SCRIPT, and we want
+  // to be able to take advantage of that in the GC hazard analysis.
+  if (!info->IsReflectable()) {
+    return NS_ERROR_FAILURE;
+  }
+
   Value* sp = nullptr;
   Value* argv = nullptr;
   uint8_t i;
@@ -765,6 +786,9 @@ nsXPCWrappedJS::CallMethod(uint16_t methodIndex, const nsXPTMethodInfo* info,
   // definitely will be when we turn off XPConnect for the web.
   RootedObject obj(RootingCx(), GetJSObject());
   nsIGlobalObject* nativeGlobal = NativeGlobal(js::UncheckedUnwrap(obj));
+
+  AutoAllowLegacyScriptExecution exemption;
+
   AutoEntryScript aes(nativeGlobal, "XPCWrappedJS method call",
                       /* aIsMainThread = */ true);
   XPCCallContext ccx(aes.cx());
@@ -774,25 +798,10 @@ nsXPCWrappedJS::CallMethod(uint16_t methodIndex, const nsXPTMethodInfo* info,
 
   JSContext* cx = ccx.GetJSContext();
 
-  if (!cx || !info->IsReflectable()) {
+  if (!cx) {
     return NS_ERROR_FAILURE;
   }
 
-  const nsXPTInterfaceInfo* interfaceInfo = GetInfo();
-  JS::RootedId id(cx);
-  const char* name;
-  nsAutoCString symbolName;
-  if (info->IsSymbol()) {
-    info->GetSymbolDescription(cx, symbolName);
-    name = symbolName.get();
-    id = SYMBOL_TO_JSID(info->GetSymbol(cx));
-  } else {
-    name = info->GetName();
-    if (!AtomizeAndPinJSString(cx, id.get(), name)) {
-      return NS_ERROR_FAILURE;
-    }
-  }
-
   // We now need to enter the realm of the actual JSObject* we are pointing at.
   // But that may be a cross-compartment wrapper and therefore not have a
   // well-defined realm, so enter the realm of the global that we grabbed back
@@ -800,6 +809,13 @@ nsXPCWrappedJS::CallMethod(uint16_t methodIndex, const nsXPTMethodInfo* info,
   RootedObject scope(cx, GetJSObjectGlobal());
   JSAutoRealm ar(cx, scope);
 
+  const nsXPTInterfaceInfo* interfaceInfo = GetInfo();
+  JS::RootedId id(cx);
+  const char* name = info->NameOrDescription();
+  if (!info->GetId(cx, id.get())) {
+    return NS_ERROR_FAILURE;
+  }
+
   // [optional_argc] has a different calling convention, which we don't
   // support for JS-implemented components.
   if (info->WantsOptArgc()) {
@@ -896,8 +912,8 @@ nsXPCWrappedJS::CallMethod(uint16_t methodIndex, const nsXPTMethodInfo* info,
     uint32_t array_count;
     RootedValue val(cx, NullValue());
 
-    // verify that null was not passed for 'out' param
-    if (param.IsOut() && !nativeParams[i].val.p) {
+    // Verify that null was not passed for a non-optional 'out' param.
+    if (param.IsOut() && !nativeParams[i].val.p && !param.IsOptional()) {
       retval = NS_ERROR_INVALID_ARG;
       goto pre_call_clean_up;
     }
@@ -1002,7 +1018,7 @@ pre_call_clean_up:
   for (i = 0; i < paramCount; i++) {
     const nsXPTParamInfo& param = info->GetParam(i);
     MOZ_ASSERT(!param.IsShared(), "[shared] implies [noscript]!");
-    if (!param.IsOut()) {
+    if (!param.IsOut() || !nativeParams[i].val.p) {
       continue;
     }
 
@@ -1096,7 +1112,7 @@ pre_call_clean_up:
 static const JSClass XPCOutParamClass = {"XPCOutParam", 0, JS_NULL_CLASS_OPS};
 
 bool xpc::IsOutObject(JSContext* cx, JSObject* obj) {
-  return js::GetObjectJSClass(obj) == &XPCOutParamClass;
+  return JS::GetClass(obj) == &XPCOutParamClass;
 }
 
 JSObject* xpc::NewOutObject(JSContext* cx) {
