@@ -1,5 +1,5 @@
-/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
-/* vim: set ts=8 sts=4 et sw=4 tw=99: */
+/* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
+/* vim: set ts=8 sts=2 et sw=2 tw=80: */
 /* This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
@@ -10,7 +10,6 @@
 #include "js/Wrapper.h"
 
 #include "mozilla/MemoryReporting.h"
-#include "mozilla/XPTInterfaceInfoManager.h"
 #include "nsIScriptError.h"
 #include "nsPrintfCString.h"
 #include "nsPointerHashKeys.h"
@@ -68,12 +67,15 @@ bool XPCNativeMember::Resolve(XPCCallContext& ccx, XPCNativeInterface* iface,
 
   if (IsMethod()) {
     const nsXPTMethodInfo* info;
-    if (NS_FAILED(iface->GetInterfaceInfo()->GetMethodInfo(mIndex, &info)))
+    if (NS_FAILED(iface->GetInterfaceInfo()->GetMethodInfo(mIndex, &info))) {
       return false;
+    }
 
     // Note: ASSUMES that retval is last arg.
-    argc = (int)info->GetParamCount();
-    if (argc && info->GetParam((uint8_t)(argc - 1)).IsRetval()) argc--;
+    argc = (int)info->ParamCount();
+    if (info->HasRetval()) {
+      argc--;
+    }
 
     callback = XPC_WN_CallMethod;
   } else {
@@ -81,12 +83,23 @@ bool XPCNativeMember::Resolve(XPCCallContext& ccx, XPCNativeInterface* iface,
     callback = XPC_WN_GetterSetter;
   }
 
-  JSFunction* fun =
-      js::NewFunctionByIdWithReserved(ccx, callback, argc, 0, GetName());
-  if (!fun) return false;
+  jsid name = GetName();
+  JS_MarkCrossZoneId(ccx, name);
+
+  JSFunction* fun;
+  if (name.isString()) {
+    fun = js::NewFunctionByIdWithReserved(ccx, callback, argc, 0, name);
+  } else {
+    fun = js::NewFunctionWithReserved(ccx, callback, argc, 0, nullptr);
+  }
+  if (!fun) {
+    return false;
+  }
 
   JSObject* funobj = JS_GetFunctionObject(fun);
-  if (!funobj) return false;
+  if (!funobj) {
+    return false;
+  }
 
   js::SetFunctionNativeReserved(funobj, XPC_FUNCTION_NATIVE_MEMBER_SLOT,
                                 PrivateValue(this));
@@ -107,101 +120,68 @@ XPCNativeInterface::~XPCNativeInterface() {
 
 // static
 already_AddRefed<XPCNativeInterface> XPCNativeInterface::GetNewOrUsed(
-    const nsIID* iid) {
+    JSContext* cx, const nsIID* iid) {
   RefPtr<XPCNativeInterface> iface;
   XPCJSRuntime* rt = XPCJSRuntime::Get();
 
   IID2NativeInterfaceMap* map = rt->GetIID2NativeInterfaceMap();
-  if (!map) return nullptr;
+  if (!map) {
+    return nullptr;
+  }
 
   iface = map->Find(*iid);
 
-  if (iface) return iface.forget();
-
-  nsCOMPtr<nsIInterfaceInfo> info;
-  XPTInterfaceInfoManager::GetSingleton()->GetInfoForIID(iid,
-                                                         getter_AddRefs(info));
-  if (!info) return nullptr;
-
-  iface = NewInstance(info);
-  if (!iface) return nullptr;
-
-  XPCNativeInterface* iface2 = map->Add(iface);
-  if (!iface2) {
-    NS_ERROR("failed to add our interface!");
-    iface = nullptr;
-  } else if (iface2 != iface) {
-    iface = iface2;
+  if (iface) {
+    return iface.forget();
   }
 
-  return iface.forget();
+  const nsXPTInterfaceInfo* info = nsXPTInterfaceInfo::ByIID(*iid);
+  if (!info) {
+    return nullptr;
+  }
+
+  return NewInstance(cx, map, info);
 }
 
 // static
 already_AddRefed<XPCNativeInterface> XPCNativeInterface::GetNewOrUsed(
-    nsIInterfaceInfo* info) {
+    JSContext* cx, const nsXPTInterfaceInfo* info) {
   RefPtr<XPCNativeInterface> iface;
 
-  const nsIID* iid;
-  if (NS_FAILED(info->GetIIDShared(&iid)) || !iid) return nullptr;
-
   XPCJSRuntime* rt = XPCJSRuntime::Get();
 
   IID2NativeInterfaceMap* map = rt->GetIID2NativeInterfaceMap();
-  if (!map) return nullptr;
-
-  iface = map->Find(*iid);
-
-  if (iface) return iface.forget();
-
-  iface = NewInstance(info);
-  if (!iface) return nullptr;
-
-  RefPtr<XPCNativeInterface> iface2 = map->Add(iface);
-  if (!iface2) {
-    NS_ERROR("failed to add our interface!");
-    iface = nullptr;
-  } else if (iface2 != iface) {
-    iface = iface2;
+  if (!map) {
+    return nullptr;
   }
 
-  return iface.forget();
+  iface = map->Find(info->IID());
+
+  if (iface) {
+    return iface.forget();
+  }
+
+  return NewInstance(cx, map, info);
 }
 
 // static
 already_AddRefed<XPCNativeInterface> XPCNativeInterface::GetNewOrUsed(
-    const char* name) {
-  nsCOMPtr<nsIInterfaceInfo> info;
-  XPTInterfaceInfoManager::GetSingleton()->GetInfoForName(name,
-                                                          getter_AddRefs(info));
-  return info ? GetNewOrUsed(info) : nullptr;
+    JSContext* cx, const char* name) {
+  const nsXPTInterfaceInfo* info = nsXPTInterfaceInfo::ByName(name);
+  return info ? GetNewOrUsed(cx, info) : nullptr;
 }
 
 // static
-already_AddRefed<XPCNativeInterface> XPCNativeInterface::GetISupports() {
+already_AddRefed<XPCNativeInterface> XPCNativeInterface::GetISupports(
+    JSContext* cx) {
   // XXX We should optimize this to cache this common XPCNativeInterface.
-  return GetNewOrUsed(&NS_GET_IID(nsISupports));
+  return GetNewOrUsed(cx, &NS_GET_IID(nsISupports));
 }
 
 // static
 already_AddRefed<XPCNativeInterface> XPCNativeInterface::NewInstance(
-    nsIInterfaceInfo* aInfo) {
-  AutoJSContext cx;
-  static const uint16_t MAX_LOCAL_MEMBER_COUNT = 16;
-  XPCNativeMember local_members[MAX_LOCAL_MEMBER_COUNT];
-  RefPtr<XPCNativeInterface> obj;
-  XPCNativeMember* members = nullptr;
-
-  int i;
-  bool failed = false;
-  uint16_t constCount;
-  uint16_t methodCount;
-  uint16_t totalCount;
-  uint16_t realTotalCount = 0;
-  XPCNativeMember* cur;
-  RootedString str(cx);
-  RootedId interfaceName(cx);
-
+    JSContext* cx, IID2NativeInterfaceMap* aMap,
+    const nsXPTInterfaceInfo* aInfo) {
   // XXX Investigate lazy init? This is a problem given the
   // 'placement new' scheme - we need to at least know how big to make
   // the object. We might do a scan of methods to determine needed size,
@@ -209,18 +189,11 @@ already_AddRefed<XPCNativeInterface> XPCNativeInterface::NewInstance(
   // Find out how often we create these objects w/o really looking at
   // (or using) the members.
 
-  bool canScript;
-  if (NS_FAILED(aInfo->IsScriptable(&canScript)) || !canScript) return nullptr;
-
-  bool mainProcessScriptableOnly;
-  if (NS_FAILED(aInfo->IsMainProcessScriptableOnly(&mainProcessScriptableOnly)))
-    return nullptr;
-  if (mainProcessScriptableOnly && !XRE_IsParentProcess()) {
+  if (aInfo->IsMainProcessScriptableOnly() && !XRE_IsParentProcess()) {
     nsCOMPtr<nsIConsoleService> console(
         do_GetService(NS_CONSOLESERVICE_CONTRACTID));
     if (console) {
-      const char* intfNameChars;
-      aInfo->GetNameShared(&intfNameChars);
+      const char* intfNameChars = aInfo->Name();
       nsPrintfCString errorMsg("Use of %s in content process is deprecated.",
                                intfNameChars);
 
@@ -229,61 +202,54 @@ already_AddRefed<XPCNativeInterface> XPCNativeInterface::NewInstance(
       nsJSUtils::GetCallingLocation(cx, filename, &lineno, &column);
       nsCOMPtr<nsIScriptError> error(
           do_CreateInstance(NS_SCRIPTERROR_CONTRACTID));
-      error->Init(NS_ConvertUTF8toUTF16(errorMsg), filename, EmptyString(),
-                  lineno, column, nsIScriptError::warningFlag,
-                  "chrome javascript");
+      error->Init(NS_ConvertUTF8toUTF16(errorMsg), filename, u""_ns, lineno,
+                  column, nsIScriptError::warningFlag, "chrome javascript",
+                  false /* from private window */,
+                  true /* from chrome context */);
       console->LogMessage(error);
     }
   }
 
-  if (NS_FAILED(aInfo->GetMethodCount(&methodCount)) ||
-      NS_FAILED(aInfo->GetConstantCount(&constCount)))
-    return nullptr;
+  // Make sure the code below does not GC. This means we don't need to trace the
+  // PropertyKeys in the MemberVector, or the XPCNativeInterface we create
+  // before it's added to the map.
+  JS::AutoCheckCannotGC nogc;
 
-  // If the interface does not have nsISupports in its inheritance chain
-  // then we know we can't reflect its methods. However, some interfaces that
-  // are used just to reflect constants are declared this way. We need to
-  // go ahead and build the thing. But, we'll ignore whatever methods it may
-  // have.
-  if (!nsXPConnect::IsISupportsDescendant(aInfo)) methodCount = 0;
+  const uint16_t methodCount = aInfo->MethodCount();
+  const uint16_t constCount = aInfo->ConstantCount();
+  const uint16_t totalCount = methodCount + constCount;
 
-  totalCount = methodCount + constCount;
-
-  if (totalCount > MAX_LOCAL_MEMBER_COUNT) {
-    members = new XPCNativeMember[totalCount];
-    if (!members) return nullptr;
-  } else {
-    members = local_members;
-  }
+  using MemberVector =
+      mozilla::Vector<XPCNativeMember, 16, InfallibleAllocPolicy>;
+  MemberVector members;
+  MOZ_ALWAYS_TRUE(members.reserve(totalCount));
 
   // NOTE: since getters and setters share a member, we might not use all
   // of the member objects.
 
-  for (i = 0; i < methodCount; i++) {
-    const nsXPTMethodInfo* info;
-    if (NS_FAILED(aInfo->GetMethodInfo(i, &info))) {
-      failed = true;
-      break;
-    }
+  for (unsigned int i = 0; i < methodCount; i++) {
+    const nsXPTMethodInfo& info = aInfo->Method(i);
 
     // don't reflect Addref or Release
-    if (i == 1 || i == 2) continue;
-
-    if (!XPCConvert::IsMethodReflectable(*info)) continue;
-
-    str = JS_AtomizeAndPinString(cx, info->GetName());
-    if (!str) {
-      NS_ERROR("bad method name");
-      failed = true;
-      break;
+    if (i == 1 || i == 2) {
+      continue;
     }
-    jsid name = INTERNED_STRING_TO_JSID(cx, str);
 
-    if (info->IsSetter()) {
-      MOZ_ASSERT(realTotalCount, "bad setter");
+    if (!info.IsReflectable()) {
+      continue;
+    }
+
+    jsid name;
+    if (!info.GetId(cx, name)) {
+      NS_ERROR("bad method name");
+      return nullptr;
+    }
+
+    if (info.IsSetter()) {
+      MOZ_ASSERT(!members.empty(), "bad setter");
       // Note: ASSUMES Getter/Setter pairs are next to each other
       // This is a rule of the typelib spec.
-      cur = &members[realTotalCount - 1];
+      XPCNativeMember* cur = &members.back();
       MOZ_ASSERT(cur->GetName() == name, "bad setter");
       MOZ_ASSERT(cur->IsReadOnlyAttribute(), "bad setter");
       MOZ_ASSERT(cur->GetIndex() == i - 1, "bad setter");
@@ -291,82 +257,87 @@ already_AddRefed<XPCNativeInterface> XPCNativeInterface::NewInstance(
     } else {
       // XXX need better way to find dups
       // MOZ_ASSERT(!LookupMemberByID(name),"duplicate method name");
-      if (realTotalCount == XPCNativeMember::GetMaxIndexInInterface()) {
+      size_t indexInInterface = members.length();
+      if (indexInInterface == XPCNativeMember::GetMaxIndexInInterface()) {
         NS_WARNING("Too many members in interface");
-        failed = true;
-        break;
+        return nullptr;
       }
-      cur = &members[realTotalCount];
-      cur->SetName(name);
-      if (info->IsGetter())
-        cur->SetReadOnlyAttribute(i);
-      else
-        cur->SetMethod(i);
-      cur->SetIndexInInterface(realTotalCount);
-      ++realTotalCount;
+      XPCNativeMember cur;
+      cur.SetName(name);
+      if (info.IsGetter()) {
+        cur.SetReadOnlyAttribute(i);
+      } else {
+        cur.SetMethod(i);
+      }
+      cur.SetIndexInInterface(indexInInterface);
+      members.infallibleAppend(cur);
     }
   }
 
-  if (!failed) {
-    for (i = 0; i < constCount; i++) {
-      RootedValue constant(cx);
-      nsCString namestr;
-      if (NS_FAILED(aInfo->GetConstant(i, &constant, getter_Copies(namestr)))) {
-        failed = true;
-        break;
-      }
-
-      str = JS_AtomizeAndPinString(cx, namestr.get());
-      if (!str) {
-        NS_ERROR("bad constant name");
-        failed = true;
-        break;
-      }
-      jsid name = INTERNED_STRING_TO_JSID(cx, str);
-
-      // XXX need better way to find dups
-      // MOZ_ASSERT(!LookupMemberByID(name),"duplicate method/constant name");
-      if (realTotalCount == XPCNativeMember::GetMaxIndexInInterface()) {
-        NS_WARNING("Too many members in interface");
-        failed = true;
-        break;
-      }
-      cur = &members[realTotalCount];
-      cur->SetName(name);
-      cur->SetConstant(i);
-      cur->SetIndexInInterface(realTotalCount);
-      ++realTotalCount;
+  for (unsigned int i = 0; i < constCount; i++) {
+    RootedValue constant(cx);
+    nsCString namestr;
+    if (NS_FAILED(aInfo->GetConstant(i, &constant, getter_Copies(namestr)))) {
+      return nullptr;
     }
+
+    RootedString str(cx, JS_AtomizeString(cx, namestr.get()));
+    if (!str) {
+      NS_ERROR("bad constant name");
+      return nullptr;
+    }
+    jsid name = PropertyKey::NonIntAtom(str);
+
+    // XXX need better way to find dups
+    // MOZ_ASSERT(!LookupMemberByID(name),"duplicate method/constant name");
+    size_t indexInInterface = members.length();
+    if (indexInInterface == XPCNativeMember::GetMaxIndexInInterface()) {
+      NS_WARNING("Too many members in interface");
+      return nullptr;
+    }
+    XPCNativeMember cur;
+    cur.SetName(name);
+    cur.SetConstant(i);
+    cur.SetIndexInInterface(indexInInterface);
+    members.infallibleAppend(cur);
   }
 
-  if (!failed) {
-    const char* bytes;
-    if (NS_FAILED(aInfo->GetNameShared(&bytes)) || !bytes ||
-        nullptr == (str = JS_AtomizeAndPinString(cx, bytes))) {
-      failed = true;
-    }
-    interfaceName = INTERNED_STRING_TO_JSID(cx, str);
+  const char* bytes = aInfo->Name();
+  if (!bytes) {
+    return nullptr;
+  }
+  RootedString str(cx, JS_AtomizeString(cx, bytes));
+  if (!str) {
+    return nullptr;
   }
 
-  if (!failed) {
-    // Use placement new to create an object with the right amount of space
-    // to hold the members array
-    int size = sizeof(XPCNativeInterface);
-    if (realTotalCount > 1)
-      size += (realTotalCount - 1) * sizeof(XPCNativeMember);
-    void* place = new char[size];
-    if (place) obj = new (place) XPCNativeInterface(aInfo, interfaceName);
+  RootedId interfaceName(cx, PropertyKey::NonIntAtom(str));
 
-    if (obj) {
-      obj->mMemberCount = realTotalCount;
-      // copy valid members
-      if (realTotalCount)
-        memcpy(obj->mMembers, members,
-               realTotalCount * sizeof(XPCNativeMember));
-    }
+  // Use placement new to create an object with the right amount of space
+  // to hold the members array
+  size_t size = sizeof(XPCNativeInterface);
+  if (members.length() > 1) {
+    size += (members.length() - 1) * sizeof(XPCNativeMember);
+  }
+  void* place = new char[size];
+  if (!place) {
+    return nullptr;
   }
 
-  if (members && members != local_members) delete[] members;
+  RefPtr<XPCNativeInterface> obj =
+      new (place) XPCNativeInterface(aInfo, interfaceName);
+
+  obj->mMemberCount = members.length();
+  // copy valid members
+  if (!members.empty()) {
+    memcpy(obj->mMembers, members.begin(),
+           members.length() * sizeof(XPCNativeMember));
+  }
+
+  if (!aMap->AddNew(obj)) {
+    NS_ERROR("failed to add our interface!");
+    return nullptr;
+  }
 
   return obj.forget();
 }
@@ -381,14 +352,29 @@ size_t XPCNativeInterface::SizeOfIncludingThis(MallocSizeOf mallocSizeOf) {
   return mallocSizeOf(this);
 }
 
+void XPCNativeInterface::Trace(JSTracer* trc) {
+  JS::TraceRoot(trc, &mName, "XPCNativeInterface::mName");
+
+  for (size_t i = 0; i < mMemberCount; i++) {
+    JS::PropertyKey key = mMembers[i].GetName();
+    JS::TraceRoot(trc, &key, "XPCNativeInterface::mMembers");
+    MOZ_ASSERT(mMembers[i].GetName() == key);
+  }
+}
+
+void IID2NativeInterfaceMap::Trace(JSTracer* trc) {
+  for (Map::Enum e(mMap); !e.empty(); e.popFront()) {
+    XPCNativeInterface* iface = e.front().value();
+    iface->Trace(trc);
+  }
+}
+
 void XPCNativeInterface::DebugDump(int16_t depth) {
 #ifdef DEBUG
-  depth--;
   XPC_LOG_ALWAYS(("XPCNativeInterface @ %p", this));
   XPC_LOG_INDENT();
   XPC_LOG_ALWAYS(("name is %s", GetNameString()));
-  XPC_LOG_ALWAYS(("mMemberCount is %d", mMemberCount));
-  XPC_LOG_ALWAYS(("mInfo @ %p", mInfo.get()));
+  XPC_LOG_ALWAYS(("mInfo @ %p", mInfo));
   XPC_LOG_OUTDENT();
 #endif
 }
@@ -396,30 +382,29 @@ void XPCNativeInterface::DebugDump(int16_t depth) {
 /***************************************************************************/
 // XPCNativeSetKey
 
-static PLDHashNumber HashPointer(const void* ptr) {
-  return nsPtrHashKey<const void>::HashKey(ptr);
-}
-
-PLDHashNumber XPCNativeSetKey::Hash() const {
-  PLDHashNumber h = 0;
+HashNumber XPCNativeSetKey::Hash() const {
+  HashNumber h = 0;
 
   if (mBaseSet) {
+    // If we ever start using mCx here, adjust the constructors accordingly.
     XPCNativeInterface** current = mBaseSet->GetInterfaceArray();
     uint16_t count = mBaseSet->GetInterfaceCount();
     for (uint16_t i = 0; i < count; i++) {
-      h ^= HashPointer(*(current++));
+      h = AddToHash(h, *(current++));
     }
   } else {
     // A newly created set will contain nsISupports first...
-    RefPtr<XPCNativeInterface> isupp = XPCNativeInterface::GetISupports();
-    h ^= HashPointer(isupp);
+    RefPtr<XPCNativeInterface> isupp = XPCNativeInterface::GetISupports(mCx);
+    h = AddToHash(h, isupp.get());
 
     // ...but no more than once.
-    if (isupp == mAddition) return h;
+    if (isupp == mAddition) {
+      return h;
+    }
   }
 
   if (mAddition) {
-    h ^= HashPointer(mAddition);
+    h = AddToHash(h, mAddition.get());
   }
 
   return h;
@@ -431,6 +416,7 @@ PLDHashNumber XPCNativeSetKey::Hash() const {
 XPCNativeSet::~XPCNativeSet() {
   // Remove |this| before we clear the interfaces to ensure that the
   // hashtable look up is correct.
+
   XPCJSRuntime::Get()->GetNativeSetMap()->Remove(this);
 
   for (int i = 0; i < mInterfaceCount; i++) {
@@ -439,22 +425,31 @@ XPCNativeSet::~XPCNativeSet() {
 }
 
 // static
-already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(const nsIID* iid) {
-  RefPtr<XPCNativeInterface> iface = XPCNativeInterface::GetNewOrUsed(iid);
-  if (!iface) return nullptr;
+already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(JSContext* cx,
+                                                          const nsIID* iid) {
+  RefPtr<XPCNativeInterface> iface = XPCNativeInterface::GetNewOrUsed(cx, iid);
+  if (!iface) {
+    return nullptr;
+  }
 
-  XPCNativeSetKey key(iface);
+  XPCNativeSetKey key(cx, iface);
 
   XPCJSRuntime* xpcrt = XPCJSRuntime::Get();
   NativeSetMap* map = xpcrt->GetNativeSetMap();
-  if (!map) return nullptr;
+  if (!map) {
+    return nullptr;
+  }
 
   RefPtr<XPCNativeSet> set = map->Find(&key);
 
-  if (set) return set.forget();
+  if (set) {
+    return set.forget();
+  }
 
-  set = NewInstance({iface.forget()});
-  if (!set) return nullptr;
+  set = NewInstance(cx, {std::move(iface)});
+  if (!set) {
+    return nullptr;
+  }
 
   if (!map->AddNew(&key, set)) {
     NS_ERROR("failed to add our set!");
@@ -466,78 +461,64 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(const nsIID* iid) {
 
 // static
 already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
-    nsIClassInfo* classInfo) {
+    JSContext* cx, nsIClassInfo* classInfo) {
   XPCJSRuntime* xpcrt = XPCJSRuntime::Get();
   ClassInfo2NativeSetMap* map = xpcrt->GetClassInfo2NativeSetMap();
-  if (!map) return nullptr;
+  if (!map) {
+    return nullptr;
+  }
 
   RefPtr<XPCNativeSet> set = map->Find(classInfo);
 
-  if (set) return set.forget();
+  if (set) {
+    return set.forget();
+  }
 
-  nsIID** iidArray = nullptr;
-  uint32_t iidCount = 0;
-
-  if (NS_FAILED(classInfo->GetInterfaces(&iidCount, &iidArray))) {
+  AutoTArray<nsIID, 4> iids;
+  if (NS_FAILED(classInfo->GetInterfaces(iids))) {
     // Note: I'm making it OK for this call to fail so that one can add
     // nsIClassInfo to classes implemented in script without requiring this
     // method to be implemented.
 
     // Make sure these are set correctly...
-    iidArray = nullptr;
-    iidCount = 0;
+    iids.Clear();
   }
 
-  MOZ_ASSERT((iidCount && iidArray) || !(iidCount || iidArray),
-             "GetInterfaces returned bad array");
-
-  // !!! from here on we only exit through the 'out' label !!!
-
-  if (iidCount) {
-    nsTArray<RefPtr<XPCNativeInterface>> interfaceArray(iidCount);
-    nsIID** currentIID = iidArray;
-
-    for (uint32_t i = 0; i < iidCount; i++) {
-      nsIID* iid = *(currentIID++);
-      if (!iid) {
-        NS_ERROR("Null found in classinfo interface list");
-        continue;
-      }
-
-      RefPtr<XPCNativeInterface> iface = XPCNativeInterface::GetNewOrUsed(iid);
-
-      if (!iface) {
-        // XXX warn here
-        continue;
-      }
-
-      interfaceArray.AppendElement(iface.forget());
+  // Try to look up each IID's XPCNativeInterface object.
+  nsTArray<RefPtr<XPCNativeInterface>> interfaces(iids.Length());
+  for (auto& iid : iids) {
+    RefPtr<XPCNativeInterface> iface =
+        XPCNativeInterface::GetNewOrUsed(cx, &iid);
+    if (iface) {
+      interfaces.AppendElement(iface.forget());
     }
+  }
 
-    if (interfaceArray.Length() > 0) {
-      set = NewInstance(Move(interfaceArray));
-      if (set) {
-        NativeSetMap* map2 = xpcrt->GetNativeSetMap();
-        if (!map2) goto out;
-
-        XPCNativeSetKey key(set);
-
-        XPCNativeSet* set2 = map2->Add(&key, set);
-        if (!set2) {
-          NS_ERROR("failed to add our set!");
-          set = nullptr;
-          goto out;
-        }
-        // It is okay to find an existing entry here because
-        // we did not look for one before we called Add().
-        if (set2 != set) {
-          set = set2;
-        }
+  // Build a set from the interfaces specified here.
+  if (interfaces.Length() > 0) {
+    set = NewInstance(cx, std::move(interfaces));
+    if (set) {
+      NativeSetMap* map2 = xpcrt->GetNativeSetMap();
+      if (!map2) {
+        return set.forget();
       }
-    } else
-      set = GetNewOrUsed(&NS_GET_IID(nsISupports));
-  } else
-    set = GetNewOrUsed(&NS_GET_IID(nsISupports));
+
+      XPCNativeSetKey key(set);
+      XPCNativeSet* set2 = map2->Add(&key, set);
+      if (!set2) {
+        NS_ERROR("failed to add our set");
+        return nullptr;
+      }
+
+      // It is okay to find an existing entry here because
+      // we did not look for one before we called Add().
+      if (set2 != set) {
+        set = set2;
+      }
+    }
+  } else {
+    set = GetNewOrUsed(cx, &NS_GET_IID(nsISupports));
+  }
 
   if (set) {
 #ifdef DEBUG
@@ -548,9 +529,6 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
     MOZ_ASSERT(set2 == set, "hashtables inconsistent!");
   }
 
-out:
-  if (iidArray) NS_FREE_XPCOM_ALLOCATED_POINTER_ARRAY(iidCount, iidArray);
-
   return set.forget();
 }
 
@@ -558,25 +536,34 @@ out:
 void XPCNativeSet::ClearCacheEntryForClassInfo(nsIClassInfo* classInfo) {
   XPCJSRuntime* xpcrt = nsXPConnect::GetRuntimeInstance();
   ClassInfo2NativeSetMap* map = xpcrt->GetClassInfo2NativeSetMap();
-  if (map) map->Remove(classInfo);
+  if (map) {
+    map->Remove(classInfo);
+  }
 }
 
 // static
 already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
-    XPCNativeSetKey* key) {
+    JSContext* cx, XPCNativeSetKey* key) {
   NativeSetMap* map = XPCJSRuntime::Get()->GetNativeSetMap();
-  if (!map) return nullptr;
+  if (!map) {
+    return nullptr;
+  }
 
   RefPtr<XPCNativeSet> set = map->Find(key);
 
-  if (set) return set.forget();
+  if (set) {
+    return set.forget();
+  }
 
-  if (key->GetBaseSet())
+  if (key->GetBaseSet()) {
     set = NewInstanceMutate(key);
-  else
-    set = NewInstance({key->GetAddition()});
+  } else {
+    set = NewInstance(cx, {key->GetAddition()});
+  }
 
-  if (!set) return nullptr;
+  if (!set) {
+    return nullptr;
+  }
 
   if (!map->AddNew(key, set)) {
     NS_ERROR("failed to add our set!");
@@ -588,23 +575,27 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
 
 // static
 already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
-    XPCNativeSet* firstSet, XPCNativeSet* secondSet,
+    JSContext* cx, XPCNativeSet* firstSet, XPCNativeSet* secondSet,
     bool preserveFirstSetOrder) {
   // Figure out how many interfaces we'll need in the new set.
   uint32_t uniqueCount = firstSet->mInterfaceCount;
   for (uint32_t i = 0; i < secondSet->mInterfaceCount; ++i) {
-    if (!firstSet->HasInterface(secondSet->mInterfaces[i])) uniqueCount++;
+    if (!firstSet->HasInterface(secondSet->mInterfaces[i])) {
+      uniqueCount++;
+    }
   }
 
   // If everything in secondSet was a duplicate, we can just use the first
   // set.
-  if (uniqueCount == firstSet->mInterfaceCount)
+  if (uniqueCount == firstSet->mInterfaceCount) {
     return RefPtr<XPCNativeSet>(firstSet).forget();
+  }
 
   // If the secondSet is just a superset of the first, we can use it provided
   // that the caller doesn't care about ordering.
-  if (!preserveFirstSetOrder && uniqueCount == secondSet->mInterfaceCount)
+  if (!preserveFirstSetOrder && uniqueCount == secondSet->mInterfaceCount) {
     return RefPtr<XPCNativeSet>(secondSet).forget();
+  }
 
   // Ok, darn. Now we have to make a new set.
   //
@@ -619,8 +610,10 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
     if (!currentSet->HasInterface(iface)) {
       // Create a new augmented set, inserting this interface at the end.
       XPCNativeSetKey key(currentSet, iface);
-      currentSet = XPCNativeSet::GetNewOrUsed(&key);
-      if (!currentSet) return nullptr;
+      currentSet = XPCNativeSet::GetNewOrUsed(cx, &key);
+      if (!currentSet) {
+        return nullptr;
+      }
     }
   }
 
@@ -631,41 +624,46 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::GetNewOrUsed(
 
 // static
 already_AddRefed<XPCNativeSet> XPCNativeSet::NewInstance(
-    nsTArray<RefPtr<XPCNativeInterface>>&& array) {
-  if (array.Length() == 0) return nullptr;
+    JSContext* cx, nsTArray<RefPtr<XPCNativeInterface>>&& array) {
+  if (array.Length() == 0) {
+    return nullptr;
+  }
 
   // We impose the invariant:
   // "All sets have exactly one nsISupports interface and it comes first."
   // This is the place where we impose that rule - even if given inputs
   // that don't exactly follow the rule.
 
-  RefPtr<XPCNativeInterface> isup = XPCNativeInterface::GetISupports();
+  RefPtr<XPCNativeInterface> isup = XPCNativeInterface::GetISupports(cx);
   uint16_t slots = array.Length() + 1;
 
   for (auto key = array.begin(); key != array.end(); key++) {
-    if (*key == isup) slots--;
+    if (*key == isup) {
+      slots--;
+    }
   }
 
   // Use placement new to create an object with the right amount of space
   // to hold the members array
   int size = sizeof(XPCNativeSet);
-  if (slots > 1) size += (slots - 1) * sizeof(XPCNativeInterface*);
+  if (slots > 1) {
+    size += (slots - 1) * sizeof(XPCNativeInterface*);
+  }
   void* place = new char[size];
   RefPtr<XPCNativeSet> obj = new (place) XPCNativeSet();
 
   // Stick the nsISupports in front and skip additional nsISupport(s)
   XPCNativeInterface** outp = (XPCNativeInterface**)&obj->mInterfaces;
-  uint16_t memberCount = 1;  // for the one member in nsISupports
 
   NS_ADDREF(*(outp++) = isup);
 
   for (auto key = array.begin(); key != array.end(); key++) {
-    RefPtr<XPCNativeInterface> cur = key->forget();
-    if (isup == cur) continue;
-    memberCount += cur->GetMemberCount();
+    RefPtr<XPCNativeInterface> cur = std::move(*key);
+    if (isup == cur) {
+      continue;
+    }
     *(outp++) = cur.forget().take();
   }
-  obj->mMemberCount = memberCount;
   obj->mInterfaceCount = slots;
 
   return obj.forget();
@@ -679,7 +677,9 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::NewInstanceMutate(
 
   MOZ_ASSERT(otherSet);
 
-  if (!newInterface) return nullptr;
+  if (!newInterface) {
+    return nullptr;
+  }
 
   // Use placement new to create an object with the right amount of space
   // to hold the members array
@@ -688,8 +688,6 @@ already_AddRefed<XPCNativeSet> XPCNativeSet::NewInstanceMutate(
   void* place = new char[size];
   RefPtr<XPCNativeSet> obj = new (place) XPCNativeSet();
 
-  obj->mMemberCount =
-      otherSet->GetMemberCount() + newInterface->GetMemberCount();
   obj->mInterfaceCount = otherSet->mInterfaceCount + 1;
 
   XPCNativeInterface** src = otherSet->mInterfaces;
@@ -720,10 +718,10 @@ void XPCNativeSet::DebugDump(int16_t depth) {
 
   XPC_LOG_ALWAYS(("mInterfaceCount of %d", mInterfaceCount));
   if (depth) {
-    for (uint16_t i = 0; i < mInterfaceCount; i++)
+    for (uint16_t i = 0; i < mInterfaceCount; i++) {
       mInterfaces[i]->DebugDump(depth);
+    }
   }
-  XPC_LOG_ALWAYS(("mMemberCount of %d", mMemberCount));
   XPC_LOG_OUTDENT();
 #endif
 }
